# AddOn RE:Group (c) 2013 Atom
# RE:Group 1.0.
# Advanced group manager that offsets animation in groups from group participants.
# Turns animated groups into offset driven clusters.
# Last Revision 07-25-2013
# Last Revision 07-10-2013

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
import os,sys,colorsys

import mathutils
from mathutils import Vector, Matrix

import random
from random import randint, uniform

# Node color logic.
COLOR_LINK_MULT = 1.0
COLOR_UNLINK_MULT = 0.42
COLOR_INT = (0.7, 1.0, 0.8, 1.0)
COLOR_FLOAT = (0.6, 0.0, 0.8, 1.0)
COLOR_STRING = (0.279, 1.0, 0.009, 1.0)
COLOR_LIST = (1.0, 0.824, 0.0, 1.0)

#####################################################################
# Globals.
#####################################################################
# Global busy flag. Try to avoid events if we are already busy.
isBusy = False
isRendering = False
lastFrameUpdated = 0.0
frameChangeCount = 0
	  
# Objects are managed by name prefix. Customize here...
OBJECT_PREFIX = "Blendgraph_"      #For master control object naming.
BLENDGRAPH_OB_PREFIX = "bg_"       #For managed object naming.
BLENDGRAPH_MESH_PREFIX = "me_bg_"  #For managed datablock naming.
BLENDGRAPH_REFERENCE_PREFIX = "rbg_"  	#For managed datablock naming.
MATERIAL_PREFIX = "mat_bg_"
ENTRY_NAME = "Blendgraph"			#For new list entries.
				
MAX_NAME_SIZE = 52
GLOBAL_ZERO_PADDING = 3             # The number of zeros to pad strings with when converting INTs to STRINGs.
DELIMITER = ","

#####################################################################
# Simple debug message control.
#####################################################################
SHOW_MESSAGES = True
MSG_PREFIX = "bg: "
def to_console(passedItem):
	if SHOW_MESSAGES == True:
		if len(passedItem) == 0:
			print("")
		else:
			s = str(passedItem)
			print(MSG_PREFIX + s)

#####################################################################
# Node update.
#####################################################################
def updateNode(self,context):
	updated = [self.name]
	self.update()
	for output in self.outputs:
		if output.is_linked:
			for link in output.links:
				nod = link.to_socket.node
				if nod.name not in updated:
					nod.update()
					updated.append(nod.name)
	del updated
	
#####################################################################
# Memory Management.
#####################################################################
def removeCurveFromMemory (passedName):
	# Extra test because this can crash Blender if not done correctly.
	result = False
	curve = bpy.data.curves.get(passedName)
	if curve != None:
		if curve.users == 0:
			try:
				curve.user_clear()
				can_continue = True
			except:
				can_continue = False
			
			if can_continue == True:
				try:
					bpy.data.curves.remove(curve)
					result = True
					#to_console("removeCurveFromMemory: MESH [" + passedName + "] removed from memory.")
				except:
					result = False
					to_console("removeCurveFromMemory: FAILED to remove [" + passedName + "] from memory.")
			else:
				# Unable to clear users, something is holding a reference to it.
				# Can't risk removing. Favor leaving it in memory instead of risking a crash.
				to_console("removeCurveFromMemory: Unable to clear users for MESH, something is holding a reference to it.")
				result = False
		else:
			to_console ("removeCurveFromMemory: Unable to remove CURVE because it still has [" + str(curve.users) + "] users.")
	else:
		# We could not fetch it, it does not exist in memory, essentially removed.
		result = True
	return result
	
def removeMeshFromMemory (passedName):
	to_console("removeMeshFromMemory:[%s]." % passedName)
	# Extra test because this can crash Blender if not done correctly.
	result = False
	mesh = bpy.data.meshes.get(passedName)
	if mesh != None:
		if mesh.users == 0:
			try:
				mesh.user_clear()
				can_continue = True
			except:
				can_continue = False
			
			if can_continue == True:
				try:
					bpy.data.meshes.remove(mesh)
					result = True
					to_console("removeMeshFromMemory: MESH [" + passedName + "] removed from memory.")
				except:
					result = False
					to_console("removeMeshFromMemory: FAILED to remove [" + passedName + "] from memory.")
			else:
				# Unable to clear users, something is holding a reference to it.
				# Can't risk removing. Favor leaving it in memory instead of risking a crash.
				to_console("removeMeshFromMemory: Unable to clear users for MESH, something is holding a reference to it.")
				result = False
		else:
			to_console ("removeMeshFromMemory: Unable to remove MESH because it still has [" + str(mesh.users) + "] users.")
	else:
		# We could not fetch it, it does not exist in memory, essentially removed.
		to_console("We could not fetch MESH [%s], it does not exist in memory, essentially removed." % passedName)
		result = True
	return result

def removeObjectFromMemory (passedName):
	to_console("removeObjectFromMemory:[%s]." % passedName)
	# Extra test because this can crash Blender if not done correctly.
	result = False
	ob = bpy.data.objects.get(passedName)
	if ob != None:
		if ob.users == 0:
			try:
				ob.user_clear()
				can_continue = True
			except:
				can_continue = False
			
			if can_continue == True:
				try:
					bpy.data.objects.remove(ob)
					result = True
					to_console("removeObjectFromMemory: OB [" + passedName + "] removed from memory.")
				except:
					result = False
					to_console("removeObjectFromMemory: FAILED to remove OB [" + passedName + "] from memory.")
			else:
				# Unable to clear users, something is holding a reference to it.
				# Can't risk removing. Favor leaving it in memory instead of risking a crash.
				to_console("removeObjectFromMemory: Unable to clear users for OBJECT, something is holding a reference to it.")
				result = False
		else:
			to_console ("removeObjectFromMemory: Unable to remove OBJECT because it still has [" + str(ob.users) + "] users.") 
	else:
		# We could not fetch it, it does not exist in memory, essentially removed.
		to_console("We could not fetch OB [%s], it does not exist in memory, essentially removed." % passedName)
		result = True
	return result 

def eraseAllKeyframes(scene,passedOB = None):
	if passedOB != None:
		ad = passedOB.animation_data
		if ad != None:
			passedOB.animation_data_clear()

#####################################################################
# Curve Management
#####################################################################
def curveFromEndPoints(curveName = "lerp", v1=Vector((0,0,0)),v2=Vector((0,0,1)),entry=None):
	subdiv = entry.taper_point_count
	noise = entry.noise
	threshold = entry.threshold
	
	v_sliced = [v1] + [v1.lerp(v2, i/subdiv) for i in range(1, subdiv + 1)]
	
	# Now make a curve out of the new vertices.
	curve_datablock_name = "cu_%s" % curveName
	cu_datablock = bpy.data.curves.get(curve_datablock_name)
	if cu_datablock == None:
		cu_datablock = bpy.data.curves.new(curve_datablock_name,'CURVE')    # Create a new curve.
		cu_datablock.dimensions = '2D'
	else:
		#Reuse this existing curve, but we must remove all of the current splines.
		for old_spline in cu_datablock.splines:
			try:
				cu_datablock.splines.remove(old_spline)
			except:
				to_console("curveFromEndPoints: ERROR trying to remove old_spline")
	
	curve_name = curveName
	ob_curve = bpy.data.objects.get(curve_name)
	if ob_curve == None:
		try:    
			ob_curve = bpy.data.objects.new(curve_name,cu_datablock)                    # Create a new curve type object.
			bpy.context.scene.objects.link(ob_curve)                                    # Link it to the scene.
		except:
			to_console("curveFromEndPoints: ERROR, failed to create tracer curve to [%s]." % curve_name)
	else:
		#to_console("WARNING: [" + curve_name + "] already exists.")
		pass
		
	# Populate the points.
	spline = cu_datablock.splines.new('BEZIER')               # Add an empty spline to the curve.
	spline.bezier_points.add(len(v_sliced)-1)                   # Pre-create the number of points in this spline based upon passed value.
	for t,co in enumerate(v_sliced):
		canRandomize = True
		if t == 0: canRandomize = False
		if t == (len(v_sliced)-1): canRandomize = False
		if canRandomize == True:
			# Only randomize points inbetween the end points.
			if random.random() > threshold:
				#p.radius = random.random()
				#for i in range(3):
				#	co[i] += noise*(random.random()*2-1)
				co[1] += noise*(random.random()*entry.random_strength-1)	 
		p = spline.bezier_points[t]
		p.co = co 
		p.tilt = 0.0
		p.radius = 0.0
		p.handle_right_type='VECTOR'
		p.handle_left_type='VECTOR'
	return ob_curve
	
#####################################################################
# Name management
##################################################################### 
def returnAllObjectNames (scene):
	# NOTE: This returns all object names in the passed scene.
	result = []
	for ob in scene.objects:
		result.append(ob.name)
	return result 

def returnObjectNamesLike(passedScene, passedName):
	# Return objects named like our passedName in the passed scene.
	result = []
	isLike = passedName
	l = len(isLike)
	all_obs = returnAllObjectNames(passedScene)
	for name in all_obs:
		candidate = name[0:l]
		if isLike == candidate:
			result.append(name)
	return result

def returnNameForNumber(passedFrame):
	frame_number = str(passedFrame)
	post_fix = frame_number.zfill(GLOBAL_ZERO_PADDING)
	return post_fix

def returnNameDroppedPrefix(ob):
	# Use only right half of name, i.e. discard xxxx_ prefix.
	lst_name = ob.name.split('_')
	return lst_name[1]
   
############################################################################
# Blender Internal Material code.
############################################################################
def returnMatteMaterial (passedName,passedRGB):
	tempMat = bpy.data.materials.get(passedName)
	if tempMat == None:
		tempMat = bpy.data.materials.new(passedName)
	tempMat.diffuse_color = (passedRGB[0],passedRGB[1],passedRGB[2])
	tempMat.specular_intensity = 0.0
	#tempMat.specular_hardness = 96
	return tempMat

def returnGlassMaterial (passedName,passedRGB):
	tempMat = bpy.data.materials.get(passedName)
	if tempMat == None:
		tempMat = bpy.data.materials.new(passedName)
	tempMat.use_transparency = True
	tempMat.transparency_method = 'RAYTRACE'
	tempMat.alpha = 0.1
	tempMat.raytrace_transparency.ior = 1.15 + randFloat(0.0,0.72)    # IOR Crown Glass is 1.52.
	tempMat.diffuse_color = (passedRGB[0],passedRGB[1],passedRGB[2])
	return tempMat

def returnMirrorMaterial (passedName,passedRGB):
	tempMat = bpy.data.materials.get(passedName)
	if tempMat == None:
		tempMat = bpy.data.materials.new(passedName)
	tempMat.raytrace_mirror.use = True
	tempMat.raytrace_mirror.reflect_factor = 0.7
	tempMat.diffuse_color = (passedRGB[0],passedRGB[1],passedRGB[2])
	return tempMat

def returnToonMaterial (passedName,passedRGB):
	tempMat = bpy.data.materials.get(passedName)
	if tempMat == None:
		tempMat = bpy.data.materials.new(passedName)
	tempMat.specular_intensity = 0.0
	tempMat.diffuse_color = (passedRGB[0],passedRGB[1],passedRGB[2])
	return tempMat

def returnEmissiveMaterial (passedName,passedRGB):
	tempMat = bpy.data.materials.get(passedName)
	if tempMat == None:
		tempMat = bpy.data.materials.new(passedName)
	tempMat.emit = 1.0
	tempMat.diffuse_color = (passedRGB[0],passedRGB[1],passedRGB[2])
	return tempMat

def returnShadowMatteMaterial (passedName,passedRGB):
	tempMat = bpy.data.materials.get(passedName)
	if tempMat == None:
		tempMat = bpy.data.materials.new(passedName)
	tempMat.use_only_shadow = True
	tempMat.diffuse_color = (passedRGB[0],passedRGB[1],passedRGB[2])
	return tempMat

def setMaterial(ob, mat):
	me = ob.data
	while len(me.materials) > 0:
		to_console("Removing previous material on [%s]." % ob.name)
		mat = me.materials.pop(0, update_data=True)
		print(mat.name)
	me.materials.append(mat)

############################################################################
# Random code.
############################################################################
def randFloat(a, b):
	#random.seed(RANDOM_SEED)
	return random.random()*(b-a)+a

def randInt(a, b):
	#random.seed(RANDOM_SEED)
	return random.randint(int(a),int(b))

def returnRandomColor (color_base = None):
	if color_base == None:
		# Ugly random color.
		return (randFloat(0.2,1.0),randFloat(0.2,1.0),randFloat(0.2,1.0))
	else:
		# Random color based upon base color hue.
		h, s, v = colorsys.rgb_to_hsv(color_base[0],color_base[1],color_base[2])
		# Use the HUE from color base to determine this unique random color.
		return (returnRndRGBFromHue (h))

############################################################################
# Color code.
############################################################################
def returnNumberOfColors(count):
	N = count
	HSV_tuples = [(x*1.0/N, 0.5, 0.5) for x in range(N)]
	RGB_tuples = map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples)
	return RGB_tuples

def returnRndRGBFromHue (h):
	#h = uniform(0.25, 0.38) # Select random green'ish hue from hue wheel
	s = uniform(0.4, 1.0)
	v = uniform(0.2, 0.9)
	r, g, b = colorsys.hsv_to_rgb(h, s, v)
	return r,g,b